from ..base_hardware_service import (
    BaseHardwareInterfaceService,
    AquireProjection,
    ReachabilityCheck,
)
from rq_controller.common import PyProjection


import ros2_numpy
import numpy as np

from sensor_msgs.msg import Image

try:
    from rosct_nodes.detector.client import DetectorClient, TriggerMode, AcquisitionMode  # type: ignore
    from rosct_nodes.xraysource.client import XraySourceClient  # type: ignore
    # from rosct_nodes.ct_sequence.ct_sequence_node import CtSequenceNode  # type: ignore

    from PythonTools.imageio import ImageFormat  # type: ignore
    from PythonTools.ezrt_header import AcquisitionGeometry  # type: ignore

    import rclpy
    from threading import Event
    import rosct_nodes.utilities.sequence_helper as sh  # type: ignore

    from trajectory_toolbox.sorter import turntable_path_planning  # type: ignore
    from roboct_moveit_commander_utils import RobotGoalTarget_from_Matrix  # type: ignore

    # from geometry_msgs.msg import Point, Quaternion
    from roboct_interfaces.msg import PlannerParameter  # type: ignore

    from rosct_roboct_interface.roboct_moveit_commander_client import (
        RoboCtMoveItCommander,
    )  # type: ignore
    from trajectory_toolbox.util.util_trajectory import (
        TrajectoryPoseManager,
        TrajectoryBase,
    )  # type: ignore
    from trajectory_toolbox.util.util_poses import Frame, ScanPoseList, ScanPoseTuple  # type: ignore

except ModuleNotFoundError:
    import warnings

    warnings.warn(
        "Packages are spedific for the RoboCT of the Deggendorf Institute of Technology. \n"
        + "Please Contact: simon.wittl@th-deg.de"
    )

    class Event:
        pass

    class ImageFormat:
        pass


class TurntablePlannerParameter:
    golden_angle_degree: float = 0.0
    angle_weight: float = 0.01


def transform_frames(frame1: Frame, frame2: Frame):
    matrix = frame1.transform @ frame2.transform
    return Frame.from_transform(matrix)


def frame_mm_to_m(frame: Frame) -> np.ndarray:
    transform = frame.transform
    transform[:3, 3] /= 1000.0
    return transform


def frame_m_to_mm(frame: Frame) -> np.ndarray:
    transform = frame.transform
    transform[:3, 3] *= 1000.0
    return transform


def frame_from_transform_msg(msg) -> Frame:
    return Frame.from_position_and_quaternion(
        np.array([msg.position.x, msg.position.y, msg.position.z]),
        np.array(
            [msg.orientation.x, msg.orientation.y, msg.orientation.z, msg.orientation.w]
        ),
    )


class ThdRoboCTHardwareInterface(BaseHardwareInterfaceService):
    interface_type: str = "THD"

    def __init__(self, nodename: str = "rq_thd_hardware_interface_service"):
        super().__init__(nodename)

        self.client_detector = DetectorClient(name="rq_detector_client", namespace=None)
        self.client_xraysource = XraySourceClient(
            name="rq_source_client", namespace=None
        )
        self.client_move_it_commander = RoboCtMoveItCommander()

        # check_xray_hardware
        # self.check_xray_hardware_connections()
        # self.check_xray_voltage()

        # init robot planner
        self.configure_turntable_planner_parameter()
        self.configure_planner_parameter()

    def print_info(self, msg: str):
        self.get_logger().info(msg)

    def print_error(self, msg: str):
        self.get_logger().error(msg)

    @property
    def stop_event(self) -> Event:
        return sh.get_stop_event(self.ct_node)

    def check_xray_hardware_connections(self):
        self.print_info("CHECK: XRay source and detector connections.")
        if not self.client_xraysource.is_connected:
            future = self.client_xraysource.connect()
            self.client_xraysource.wait_for_future_ready(future)

        if not self.client_detector.is_connected:
            future = self.client_detector.connect()
            self.client_detector.wait_for_future_ready(future)

    def check_xray_voltage(self):
        if not self.client_xraysource.is_xray_on:
            xray_on_future = self.client_xraysource.xray_on()
            self.client_xraysource.wait_for_future_ready(
                xray_on_future, timeout_sec=180.0
            )

        if not self.client_xraysource.poll_until_target_voltage_reached(
            self.stop_event
        ):
            self.print_error("Xray source voltage not reached!")

    def configure_detector(
        self,
        number_of_projections: str,
        installed_prefilter: str,
        image_format: ImageFormat,
    ):
        self.print_info("CONFIGURE: Detector.")

        self.client_detector.reset_image_metadata()

        # this sequences runs in software trigger mode, thus the detector settings are adjusted -
        # the old trigger mode is saved and later reset
        self.previous_trigger_mode = self.client_detector.trigger_mode
        self.client_detector.trigger_mode = TriggerMode.SOFTWARE

        measurement_name = self.measurement_name
        self.client_detector.set_image_metadata("measurement_name", measurement_name)
        self.client_detector.set_image_metadata(
            "acquisition_geometry", int(AcquisitionGeometry.ARBITRARY)
        )
        self.client_detector.set_image_metadata(
            "number_of_projections", number_of_projections
        )
        self.client_detector.set_image_metadata(
            "voltage_in_kv", self.client_xraysource.voltage_in_kv
        )
        self.client_detector.set_image_metadata(
            "current_in_ua", self.client_xraysource.current_in_ua
        )
        self.client_detector.set_image_metadata("prefilter", installed_prefilter)

        exposure_time_in_sec = self.client_detector.exposure_time_in_ms / 1000.0
        self.image_acquisition_timeout = max(
            (
                exposure_time_in_sec * self.client_detector.num_averages
                + exposure_time_in_sec * self.client_detector.num_skip_images
            )
            * 5.0,
            5.0,
        )

        # start continuous acquisition of image data
        start_acquisition_future = self.client_detector.start_acquisition(
            mode=AcquisitionMode.CONTINUOUS
        )
        self.client_detector.wait_for_future_ready(
            start_acquisition_future, timeout_sec=20.0
        )

        # initialize saving of an image series with the expected number of projections (saving is stopped after the stated
        # amount of images), the desired fileprefix, savepath and imageformat. If it is a continued CT adjust values
        # accordingly. The used target save path (e.g. with time prefix) for saving is returned for later use.
        _ = self.client_detector.save_image_series(
            number_of_projections,
            fileprefix=measurement_name,
            startindex=0,
            savepath=str(self.projection_tempfolder),
            create_prefix_dir=True,
            imageformat=image_format,
        )

        # all following images are tagged as valid projections
        self.client_detector.set_image_metadata("tag", "projection")

    def configure_planner_parameter(
        self,
        acceleration_scaling: float = 0.9,
        velocity_scaling: float = 0.9,
        number_of_parallel_plans: int = 12,
        planning_time: float = 25.0,
    ):
        """Configuration for the path planning of the robot via the MoveIt2 plugin

        Args:
            acceleration_scaling (float, optional): Maximal allowed acceleration for all joints. Defaults to 1..
            velocity_scaling (float, optional): Maximal allowed velocity for all joint. Defaults to 0.5.
            number_of_parallel_plans (int, optional): Defaults to 12.
            planning_time (float, optional): Maximal planning time before stop. Defaults to 5.
        """
        self.print_info("CONFIGURE: Planner parameters.")
        self.parameter_planner = PlannerParameter()
        self.parameter_planner.acceleration_scaling = acceleration_scaling
        self.parameter_planner.velocity_scaling = velocity_scaling
        self.parameter_planner.number_of_parallel_plans = number_of_parallel_plans
        self.parameter_planner.planning_time = planning_time

    def configure_turntable_planner_parameter(
        self, golden_angle_degree: float = 0, angle_weight: float = 1e7
    ) -> TurntablePlannerParameter:
        """Configuration for the use of the turn table to maximize the reachable scan poses.

        Args:
            golden_angle_degree (float, optional): All Primary beams are parallel to this z-angle. Defaults to 0.
            angle_weight (float, optional): Weight for the feature vector (Position, Turntable angle * w) for the minimizing of the robot path length of a full ct trajectory. Defaults to 0.01.

        Returns:
            TurntablePlannerParameter: Class for storing the turntalbe planning parameters
        """
        self.parameter_turntable_planer = TurntablePlannerParameter()
        self.parameter_turntable_planer.golden_angle_degree = golden_angle_degree
        self.parameter_turntable_planer.angle_weight = angle_weight
        self.print_info(f"TURNTABLEPLANNING: {golden_angle_degree} {angle_weight}")

    def aquire_projection_callback(
        self, request: AquireProjection.Request, response: AquireProjection.Response
    ):
        trajectory = TrajectoryBase("rq_move")
        scan_pose_list = ScanPoseList()
        planned_projection = PyProjection.from_message(request.scan_pose)

        # get trafo world -> object
        world_object_m = self.client_move_it_commander.get_transform(
            "world", "object"
        )  # in meter !!!

        object_frame_m = frame_from_transform_msg(world_object_m)
        object_frame_mm = Frame.from_transform(frame_m_to_mm(object_frame_m))

        source_frame: Frame = Frame.from_position_and_quaternion(
            planned_projection.focal_spot_mm,
            planned_projection.focal_spot_orientation_quad,
        )
        source_frame = transform_frames(object_frame_mm, source_frame)

        detector_frame = Frame.from_position_and_quaternion(
            planned_projection.detector_postion_mm,
            planned_projection.detector_orientation_quad,
        )
        detector_frame = transform_frames(object_frame_mm, detector_frame)

        scan_pose = ScanPoseTuple(source_frame, detector_frame)
        scan_pose_list = ScanPoseList()
        scan_pose_list.add_pose_tuple(scan_pose)

        trajectory._scan_pose_list = scan_pose_list
        tpm = TrajectoryPoseManager(trajectory)

        world_table_transformation_m = frame_from_transform_msg(
            self.client_move_it_commander.get_transform("world", "tisch_static_plane")
        )
        roated_scan_poses, angles = turntable_path_planning(
            world_table_transformation_m, tpm, golden_angle=0.0
        )

        roated_scan_poses.pose_tuples[0]._source_pose
        angle_in_rad = angles[0]
        angle_in_rad = angle_in_rad - np.pi / 2

        vector = world_table_transformation_m.position
        vector[0] += np.sin(angle_in_rad * -1) * 2.0
        vector[1] += np.cos(angle_in_rad * -1) * 2.0
        vector[2] += 1.0
        matrix = np.eye(4)
        matrix[:3, 3] = vector

        target_master = RobotGoalTarget_from_Matrix(
            "all",
            "master",
            "comet_tcp",
            frame_mm_to_m(roated_scan_poses.pose_tuples[0]._source_pose),
        )
        target_slave = RobotGoalTarget_from_Matrix(
            "all",
            "slave",
            "varex_tcp",
            frame_mm_to_m(roated_scan_poses.pose_tuples[0]._detector_pose),
        )
        target_table = RobotGoalTarget_from_Matrix(
            "all", "tisch", "tisch_virtual_tcp", matrix
        )

        req, res = self.client_move_it_commander.move_robots(
            [target_table, target_master, target_slave], self.parameter_planner
        )

        if not res.movement_executed:
            response.status.success = False
            return response

        self.client_detector.take_single_image(block=True)
        image = self.client_detector.current_image.copy()

        source_m = frame_from_transform_msg(
            self.client_move_it_commander.get_transform("tisch_tcp", "comet_tcp")
        )

        detector_m = frame_from_transform_msg(
            self.client_move_it_commander.get_transform("tisch_tcp", "varex_tcp")
        )

        projection = PyProjection(
            source_m.position * 1000.0,
            detector_m.position * 1000.0,
            detector_m.quaternion,
            image,
            427.008,
            427.008,
            "tisch",
            source_frame.quaternion,
        )

        self.projection_publisher.publish(
            ros2_numpy.msgify(Image, image.astype(np.uint16), "mono16")
        )

        response.projection = projection.as_message()
        response.status.success = True
        return response

    def check_reachability_callback(
        self, request: ReachabilityCheck.Request, response: ReachabilityCheck.Response
    ):
        pass


def main(args=None):
    rclpy.init(args=args)

    minimal_service = ThdRoboCTHardwareInterface()

    executor = rclpy.executors.MultiThreadedExecutor()
    executor.add_node(minimal_service)
    executor.add_node(minimal_service.client_detector)
    executor.add_node(minimal_service.client_xraysource)
    executor.add_node(minimal_service.client_move_it_commander)
    # executor.add_node(minimal_service.transform_broadcaster)

    # executor_thread = threading.Thread(target=executor.spin, daemon=True)
    # executor_thread.start()
    # rate = minimal_service.create_rate(2)

    # try:
    #     while rclpy.ok():
    #         print('Help me body, you are my only hope')
    #         rate.sleep()
    # except KeyboardInterrupt:
    #     pass
    # rclpy.shutdown()
    # executor_thread.join()
    try:
        executor.spin()
    except KeyboardInterrupt:
        pass
    rclpy.shutdown()


if __name__ == "__main__":
    main()
